"""
Wrapper for iblatlas
======================

Provide basic plotting usage

See detail in ``iblatlas.plots.plot_scalar_on_slice``

.. seealso:: `<https://int-brain-lab.github.io/iblenv/notebooks_external/atlas_plotting_scalar_on_slice.html>`_


Example of plot the allen map with given areas
------------------------------------------------

.. code-block:: python

    import numpy as np
    from iblatlas.plots import prepare_lr_data
    from matplotlib import pyplot as plt
    from neuralib.atlas.ibl.plot import IBLAtlasPlotWrapper

    acronyms = np.array(['VPM', 'VPL', 'PO', 'LP', 'CA1', 'DG-mo'])
    ibl.plot_scalar_on_slice(acronyms, coord=-2000, plane='coronal', background='boundary', cmap='Reds')
    plt.show()


Example of plot the automerged subregions
------------------------------------------------

.. code-block:: python

    acronyms = np.array(['RSPagl', 'RSPd', 'RSPv'])
    # `Beryl` mapping merge
    ibl.plot_scalar_on_slice(acronyms, coord=-2000, plane='coronal', background='image', cmap='Reds', mapping='Beryl')
    plt.show()


Example of plot the sagittal subregions
------------------------------------------------


.. code-block:: python

    acronyms = np.array(['VPM', 'VPL', 'PO', 'LP', 'CA1', 'DG-mo', 'SSs5', 'VISa5', 'AUDv6a', 'MOp5', 'FRP5'])
    ibl.plot_scalar_on_slice(acronyms,
                             coord=-2000, plane='sagittal', mapping='Cosmos',
                             hemisphere='left', background='image', cmap='Greens')

    plt.show()


Example of plot the transverse view
------------------------------------------------

.. code-block:: python

    acronyms_lh = np.array(['LP', 'CA1'])
    values_lh = np.random.randint(0, 10, acronyms_lh.size)
    acronyms_rh = np.array(['DG-mo', 'VISa5'])
    values_rh = np.random.randint(0, 10, acronyms_rh.size)
    acronyms_lr, values_lr = prepare_lr_data(acronyms_lh, values_lh, acronyms_rh, values_rh)
    ibl.plot_scalar_on_slice(acronyms_lr, values=values_lr, coord=-2500, plane='horizontal', mapping='Allen',
                             hemisphere='both', background='image', cmap='Reds', clevels=[0, 5])

     plt.show()


Example of plot the Top view (dorsal cortex)
-----------------------------------------------

.. code-block:: python

    acronyms = np.array(['RSPagl', 'RSPd', 'RSPv'])
    ibl.plot_scalar_on_slice(acronyms, coord=-2000, plane='top', hemisphere='left',
                             background='boundary', cmap='Purples', mapping='Beryl', **kwargs)

    plt.show()






"""

from .plot import *



from __future__ import annotations

import functools
import re
from pathlib import Path
from typing import Any, TypeVar

from neuralib.util.util_type import PathLike, ArrayLike
from neuralib.util.util_verbose import fprint

__all__ = ['uglob',
           'glob_re',
           'joinn',
           'flat_ls',
           'array2str',
           'deprecated',
           'key_from_value']


def uglob(d: PathLike,
          pattern: str,
          sort: bool = True,
          is_dir: bool = False) -> Path:
    """
    Unique glob

    :param d: directory
    :param pattern: pattern string
    :param sort: if sort
    :param is_dir: only return if is a directory
    :return: unique path
    """
    if not isinstance(d, Path):
        d = Path(d)

    if not d.is_dir():
        raise ValueError(f'{d} is not a directory')

    f = list(d.glob(pattern))

    if is_dir:
        f = [ff for ff in f if ff.is_dir()]

    if sort:
        f.sort()

    if len(f) == 0:
        raise FileNotFoundError(f'{d} not have pattern: {pattern}')
    elif len(f) == 1:
        return f[0]
    else:
        raise RuntimeError(f'multiple files were found in {d} in pattern {pattern} >>> {f}')


def glob_re(pattern: str, strings: list[str]) -> list[str]:
    """find list of str element fit for re pattern"""
    return list(filter(re.compile(pattern).match, strings))


def joinn(sep: str, *part: str | None) -> str:
    """join not-None str"""
    return sep.join(str(it) for it in part if it is not None)


def flat_ls(ls: list[ArrayLike]) -> list[Any]:
    """flatten the list"""
    from itertools import chain
    return list(chain(*ls))


def array2str(x: ArrayLike, sep=' ') -> str:
    """
    covert array to str with sep.
    i.e., used in csv writing of array in a specific cell

    :param x: 1D
    :param sep:
    :return:
    """
    return sep.join(map(str, x))


def deprecated(f=None, *, reason: str = None):
    """Mark deprecated functions.

    :param f:
    :param reason:
    :return:
    """
    if reason is None:
        reason = '...'

    def _deprecated(f):
        # TODO caller sensitive?
        warned = False

        @functools.wraps(f)
        def _deprecated_func(*args, **kwargs):
            nonlocal warned
            if not warned:
                fprint(f'use Deprecated function {f.__name__} : {reason}', vtype='warning')
                warned = True

            return f(*args, **kwargs)

        if f.__doc__ is None:
            _deprecated_func.__doc__ = "DEPRECATED."
        else:
            _deprecated_func.__doc__ = "DEPRECATED. " + f.__doc__

        return _deprecated_func

    if f is None:
        return _deprecated
    else:
        return _deprecated(f)


# ============================== #

KT = TypeVar('KT')
VT = TypeVar('VT')


def key_from_value(d: dict[KT, VT], value: VT) -> KT:
    """Get dict key from dict value, supporting str, int, list, and tuple types for values."""
    matching_keys = []
    for key, val in d.items():
        if isinstance(val, (str, int)) and val == value:
            matching_keys.append(key)
        elif isinstance(val, (list, tuple)) and value in val:
            matching_keys.append(key)

    if not matching_keys:
        raise KeyError(f'Value {value} not found in the dictionary')
    if len(matching_keys) > 1:
        raise RuntimeError(f'Multiple keys found for the value {value}: {matching_keys}')

    return matching_keys[0]



import abc
from pathlib import Path
from typing import Literal, Union, Optional

import numpy as np
from bokeh.models import GlyphRenderer, ColumnDataSource
from bokeh.plotting import figure

from neuralib.bokeh_model.base import ViewComponent

__all__ = ['AbstractBrainView', 'BrainView', 'TiffBrainView', 'AtlasBrainView']


class AbstractBrainView(ViewComponent):
    data_brain_slice: ColumnDataSource
    render_brain_slice: GlyphRenderer

    def __init__(self):
        self.data_brain_slice = ColumnDataSource(data=dict(image=[], x=[], y=[], dw=[], dh=[]))
        self._width: Optional[float] = None
        self._height: Optional[float] = None

    @property
    def width(self) -> Optional[float]:
        return self._width

    @width.setter
    def width(self, value: float):
        self._width = value

    @property
    def height(self) -> Optional[float]:
        return self._height

    @height.setter
    def height(self, value: float):
        self._height = value

    def plot(self, fig: figure,
             palette='Greys256'):
        self.render_brain_slice = fig.image(
            'image', x='x', y='y', dw='dw', dh='dh',
            source=self.data_brain_slice,
            palette=palette, level="image",
        )

    @property
    @abc.abstractmethod
    def brain_image(self) -> np.ndarray:
        pass

    def update(self, x: float = 0, y: float = 0):
        brain = self.brain_image
        if brain is None:
            self.data_brain_slice.data = dict(
                image=[], dw=[], dh=[], x=[], y=[]
            )
            return

        h, w = brain.shape
        if self._width is None:
            self._width = w
        if self._height is None:
            self._height = h

        self.data_brain_slice.data = dict(
            image=[brain], dw=[self.width], dh=[self.height], x=[x], y=[y]
        )


class BrainView(AbstractBrainView):
    def __init__(self):
        super().__init__()
        self.reference: np.ndarray = None

    @property
    def brain_image(self) -> np.ndarray:
        return self.reference

    @brain_image.setter
    def brain_image(self, image: np.ndarray):
        self.reference = image


class TiffBrainView(AbstractBrainView):
    def __init__(self, file: Path):
        super().__init__()

        from PIL import Image

        self.file_path = file
        im = Image.open(file)
        im = im.convert("RGBA")
        self.reference = np.array(im, dtype=np.uint8)
        self._offset = 0

    @property
    def offset(self) -> int:
        return self._offset

    @offset.setter
    def offset(self, value: int):
        self._offset = value

    @property
    def brain_image(self) -> np.ndarray:
        return self.reference[self._offset]


class AtlasBrainView(AbstractBrainView):
    PLANE = Literal['coronal', 'sagittal', 'transverse']

    def __init__(self, source: str, check_latest=False, plane: PLANE = 'coronal'):
        super().__init__()

        from bg_atlasapi import BrainGlobeAtlas
        atlas = BrainGlobeAtlas(
            source,
            check_latest=check_latest,
        )

        self.resolution: int = atlas.resolution[0]
        assert self.resolution == atlas.resolution[1] == atlas.resolution[2]

        self.reference = atlas.reference

        self.grid_x: np.ndarray
        self.grid_y: np.ndarray
        self.plane = plane  # invoke setter
        self._offset = 0
        self._offset_h = 0
        self._offset_w = 0
        self._offset_m: Optional[np.ndarray] = None

    @property
    def n_ap(self) -> int:
        return self.reference.shape[0]

    @property
    def n_dv(self) -> int:
        return self.reference.shape[1]

    @property
    def n_ml(self) -> int:
        return self.reference.shape[2]

    @property
    def plane(self) -> PLANE:
        return self._plane

    @plane.setter
    def plane(self, value: PLANE):
        self._plane = value
        self.grid_y, self.grid_x = np.mgrid[0:self.height, 0:self.width]
        self._offset_w = 0
        self._offset_h = 0
        self._offset_m = None

    @property
    def n_frame(self) -> int:
        if self._plane == 'coronal':
            return self.n_ap
        elif self._plane == 'sagittal':
            return self.n_ml
        elif self._plane == 'transverse':
            return self.n_dv
        else:
            raise RuntimeError()

    @property
    def width_n(self) -> int:
        if self._plane == 'coronal':
            return self.n_ml
        elif self._plane == 'sagittal':
            return self.n_ap
        elif self._plane == 'transverse':
            return self.n_ml
        else:
            raise RuntimeError()

    @property
    def height_n(self) -> int:
        if self._plane == 'coronal':
            return self.n_dv
        elif self._plane == 'sagittal':
            return self.n_dv
        elif self._plane == 'transverse':
            return self.n_ap
        else:
            raise RuntimeError()

    @property
    def width(self) -> float:
        return self.width_n * self.resolution

    @property
    def height(self) -> float:
        return self.height_n * self.resolution

    @property
    def offset(self) -> int:
        return self._offset

    @offset.setter
    def offset(self, value: int):
        self._offset = value

    @property
    def offset_w(self) -> int:
        return self._offset_w

    @offset_w.setter
    def offset_w(self, value: int):
        self._offset_w = value
        self._offset_m = self.gen_offset_matrix(self._offset_w, self._offset_h)

    @property
    def offset_h(self) -> int:
        return self._offset_h

    @offset_h.setter
    def offset_h(self, value: int):
        self._offset_h = value
        self._offset_m = self.gen_offset_matrix(self._offset_w, self._offset_h)

    @property
    def offset_matrix(self) -> np.ndarray:
        if self._offset_m is None:
            self._offset_m = self.gen_offset_matrix(self._offset_w, self._offset_h)
        return self._offset + self._offset_m

    def gen_offset_matrix(self, h: int, v: int) -> np.ndarray:
        """

        :param h: horizontal plane diff to the center. right side positive.
        :param v: vertical plane diff to the center. bottom side positive.
        :return: (H, W) array
        """
        x_frame = np.round(np.linspace(-h, h, self.width_n)).astype(int)
        y_frame = np.round(np.linspace(-v, v, self.height_n)).astype(int)

        return x_frame[None, :] + y_frame[:, None]

    def brain_slice(self, offset: Union[int, np.ndarray]) -> np.ndarray:
        return self.reference[offset, self.grid_y, self.grid_x]

    @property
    def brain_image(self) -> np.ndarray:
        return self.brain_slice(self.offset_matrix)



import math
import operator
from random import randint

import numpy as np
from pyparsing import (
    CaselessKeyword,
    Regex,
    Word,
    alphas,
    alphanums,
    Literal,
    Suppress,
    Forward,
    delimitedList,
    Group,
    ParseException
)

__all__ = ['evaluate_string']

EXPR_STACK = []

BNF = None


class EvaluateStringExpressionError(ParseException):
    def __init__(self, iden: str):
        super().__init__(iden, msg=f"invalid identifier '{iden}'")
        self.iden = iden


def bnf_parser():
    """
    expop   :: '^'
    multop  :: '*' | '/'
    addop   :: '+' | '-'
    integer :: ['+' | '-'] '0'..'9'+
    atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
    factor  :: atom [ expop factor ]*
    term    :: factor [ multop factor ]*
    expr    :: term [ addop term ]*
    """

    global BNF

    if BNF is None:
        # use CaselessKeyword for e and pi, to avoid accidentally matching
        # functions that start with 'e' or 'pi' (such as 'exp'); Keyword
        # and CaselessKeyword only match whole words
        e = CaselessKeyword("E")
        pi = CaselessKeyword("PI")

        #
        fnumber = Regex(r"[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?")
        ident = Word(alphas, alphanums + "_$")

        #
        plus, minus, mult, div = map(Literal, "+-*/")
        lpar, rpar = map(Suppress, "()")
        addop = plus | minus
        multop = mult | div
        expop = Literal("^")

        #
        lesser_than = Literal("<")
        greater_than = Literal('>')
        lesser_or_equal = Literal("<=")
        greater_or_equal = Literal(">=")
        compop = lesser_or_equal | greater_or_equal | lesser_than | greater_than

        #
        expr = Forward()
        expr_list = delimitedList(Group(expr))

        # add parse action that replaces the function identifier with a (name, number of args) tuple
        def insert_fn_argcount_tuple(t):
            fn = t.pop(0)
            num_args = len(t[0])
            t.insert(0, (fn, num_args))

        fn_call = (ident + lpar - Group(expr_list) + rpar).setParseAction(
            insert_fn_argcount_tuple
        )
        atom = (
                (addop[...] | compop[...])
                + (
                        (fn_call | pi | e | fnumber | ident).setParseAction(push_first)
                        | Group(lpar + expr + rpar)
                )
        ).setParseAction(push_unary_minus)

        # by defining exponentiation as "atom [ ^ factor ]..." instead of "atom [ ^ atom ]...", we get right-to-left
        # exponents, instead of left-to-right that is, 2^3^2 = 2^(3^2), not (2^3)^2.
        factor = Forward()
        factor <<= atom + (expop + factor).setParseAction(push_first)[...]
        term = factor + (multop + factor).setParseAction(push_first)[...]
        expr <<= term + ((addop | compop) + term).setParseAction(push_first)[...]
        BNF = expr

    return BNF


def push_unary_minus(toks):
    for t in toks:
        if t == "-":
            EXPR_STACK.append("unary -")
        else:
            break


def push_first(toks):
    EXPR_STACK.append(toks[0])


def evaluate_string(s: str, **p) -> str:
    """

    :param s: evaluate string
    :param p:
    :return:
    """
    EXPR_STACK[:] = []

    try:
        bnf_parser().parseString(s, parseAll=True)
    except ParseException as e:
        raise RuntimeError(s) from e

    val = evaluate_stack(EXPR_STACK[:], p)
    return val


epsilon = 1e-12
opn = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.truediv,
    "^": operator.pow,
    "<": operator.lt,
    ">": operator.gt,
    "<=": operator.le,
    ">=": operator.ge
}

fn = {
    "sin": np.sin,
    "cos": np.cos,
    "tan": np.tan,
    "exp": np.exp,
    "abs": np.abs,
    "trunc": lambda it: int(it),
    "round": np.round,
    "randint": randint,
    "sgn": lambda a: -1 if a < -epsilon else 1 if a > epsilon else 0,
    "multiply": lambda a, b: a * b,
    "hypot": math.hypot,
    'all': np.all
}


def evaluate_stack(s, p: dict[str, float]):
    """

    :param s:
    :param p:
    :return:
    """
    op, num_args = s.pop(), 0
    if isinstance(op, tuple):
        op, num_args = op

    if op == "unary -":
        return -evaluate_stack(s, p)
    elif op in opn:
        # note: operands are pushed onto the stack in reverse order
        op2 = evaluate_stack(s, p)
        op1 = evaluate_stack(s, p)
        return opn[op](op1, op2)
    elif op == "PI":
        return math.pi  # 3.1415926535
    elif op == "E":
        return math.e  # 2.718281828
    elif op in fn:
        # note: args are pushed onto the stack in reverse order
        args = reversed([evaluate_stack(s, p) for _ in range(num_args)])
        if op == 'all':
            return fn[op](list(args))
        else:
            return fn[op](*args)
    elif op in p:
        return p[op]
    elif op[0].isalpha():
        raise EvaluateStringExpressionError(op)
    else:
        # try to evaluate as int first, then as float if int fails
        try:
            return int(op)
        except ValueError:
            return float(op)


def test(s, expected, **p) -> None:
    from colorama import Fore, Style
    EXPR_STACK[:] = []

    try:
        results = bnf_parser().parseString(s, parseAll=True)
        val = evaluate_stack(EXPR_STACK[:], p)
    except EvaluateStringExpressionError as e:
        print(s, p, Fore.RED + "FAILED eval:" + Style.RESET_ALL, str(e), EXPR_STACK)
    except ParseException as pe:
        print(s, p, Fore.RED + "FAILED parse:" + Style.RESET_ALL, str(pe))
    else:
        if val == expected:
            print(s, p, "=", val, results, "=>", EXPR_STACK, Fore.GREEN + " SUCCESS" + Style.RESET_ALL)
        else:
            print(s, p + "!!!", val, "!=", expected, results, "=>", EXPR_STACK,
                  Fore.RED + " WRONG VALUE" + Style.RESET_ALL)


def run_test():
    test("t+1", 3, t=2)
    test("9", 9)
    test("-9", -9)
    test("--9", 9)
    test("-E", -math.e)
    test("9 + 3 + 6", 9 + 3 + 6)
    test("9 + 3 / 11", 9 + 3.0 / 11)
    test("(9 + 3)", (9 + 3))
    test("(9+3) / 11", (9 + 3.0) / 11)
    test("9 - 12 - 6", 9 - 12 - 6)
    test("9 - (12 - 6)", 9 - (12 - 6))
    test("2*3.14159", 2 * 3.14159)
    test("3.1415926535*3.1415926535 / 10", 3.1415926535 * 3.1415926535 / 10)
    test("PI * PI / 10", math.pi * math.pi / 10)
    test("PI*PI/10", math.pi * math.pi / 10)
    test("PI^2", math.pi ** 2)
    test("round(PI^2)", round(math.pi ** 2))
    test("6.02E23 * 8.048", 6.02e23 * 8.048)
    test("e / 3", math.e / 3)
    test("sin(PI/2)", math.sin(math.pi / 2))
    test("10+sin(PI/4)^2", 10 + math.sin(math.pi / 4) ** 2)
    test("trunc(E)", int(math.e))
    test("trunc(-E)", int(-math.e))
    test("round(E)", round(math.e))
    test("round(-E)", round(-math.e))
    test("E^PI", math.e ** math.pi)
    test("exp(0)", 1)
    test("exp(1)", math.e)
    test("2^3^2", 2 ** 3 ** 2)
    test("(2^3)^2", (2 ** 3) ** 2)
    test("2^3+2", 2 ** 3 + 2)
    test("2^3+5", 2 ** 3 + 5)
    test("2^9", 2 ** 9)
    test("sgn(-2)", -1)
    test("sgn(0)", 0)
    test("sgn(0.1)", 1)
    test("round(E, 3)", round(math.e, 3))
    test("round(PI^2, 3)", round(math.pi ** 2, 3))
    test("sgn(cos(PI/4))", 1)
    test("sgn(cos(PI/2))", 0)
    test("sgn(cos(PI*3/4))", -1)
    test("+(sgn(cos(PI/4)))", 1)
    test("-(sgn(cos(PI/4)))", -1)
    test("hypot(3, 4)", 5)
    test("multiply(3, 7)", 21)
    test("all(1,1,1)", True)
    # test("all(1,1,1,1,1,0)", False)

    test("0<3", True)
    test("0>3", False)
    test("3>0", True)
    test("3<0", False)
    test("5*(0<3)", 5)
    test("5*(0<3)+2*(0>3)", 5)

    test("3<3", False)
    test("3<=3", True)
    test("0<=3", True)
    test("4<=3", False)
    test("3>=3", True)
    test("3>=0", True)
    test("0>=3", False)

    test("1+1<2*3", True)
    test("1+(1<2)*3", 4)
    test("(1+1<2)*3", 0)

    test("(-60+20*0)*(0<3)+(30*0)*(0>=3)", -60)
    test("(-60+20*4)*(4<3)+(30*4)*(4>=3)", 120)


def main():
    run_test()


if __name__ == '__main__':
    main()
